# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# ------------------------
# Standard Python imports
# -----------------------

from lica import StrEnum

# -------------------------
# Constant and Enumerations
# -------------------------

IDA_HEADER_LEN = 35


class StringEnum(StrEnum):
    @classmethod
    def names(cls):
        """Get the values in the order defined"""
        return [member.name for _, member in cls.__members__.items()]

    @classmethod
    def values(cls):
        """Get the values in the order defined"""
        return [member.value for _, member in cls.__members__.items()]


# IDA keywords in the comments section of the IDA file
class IDA_KEYWORDS(StringEnum):
    LICENSE = "License"
    NUM_HEADERS = "Number of header lines"
    NUM_CHANNELS = "Number of channels"
    OBSERVER = "Data supplier"
    LOCATION = "Location name"
    TIMEZONE = "Local timezone"
    POSITION = "Position"
    FOV = "Field of view"
    COVER_OFFSET = "TESS cover offset value"
    NUM_COLS = "Number of fields per line"
    ZP = "TESS zero point"
    AIM = "Measurement direction per channel"
    FILTERS = "Filters per channel"
    PHOT_NAME = "Instrument ID"


# TESS-W Data column names
# order is imporant: it is the oroder in the IDA file


class TESSW_COLS(StringEnum):
    UTC_TIME = "time"  # always 'time' for TimeSeries Astropy Class
    LOCAL_TIME = "Local Date & Time"
    BOX_TEMP = "Enclosure Temperature"
    SKY_TEMP = "Sky Temperature"
    FREQ = "Frequency"
    MAG = "MSAS"
    ZP = "ZP"
    SEQ_NUM = "Sequence Number"


# TESS-4C data column names
# order is imporant: it is the order in the IDA file


class TESS4C_COLS(StringEnum):
    UTC_TIME = "time"  # always 'time' for TimeSeries Astropy Class
    LOCAL_TIME = "Local Date & Time"
    BOX_TEMP = "Enclosure Temperature"
    SKY_TEMP = "Sky Temperature"
    FREQ1 = "Freq1"
    MAG1 = "MSAS1"
    ZP1 = "ZP1"
    FREQ2 = "Freq2"
    MAG2 = "MSAS2"
    ZP2 = "ZP2"
    FREQ3 = "Freq3"
    MAG3 = "MSAS3"
    ZP3 = "ZP3"
    FREQ4 = "Freq4"
    MAG4 = "MSAS4"
    ZP4 = "ZP4"
    SEQ_NUM = "Sequence Number"


# Additional data column names  that can appear in the Time Series


class TIMESERIES_COLS(StringEnum):
    SUN_ALT = "Sun Alt"
    SUN_AZ = "Sun Az"
    MOON_AZ = "Moon Az"
    MOON_ALT = "Moon Alt"
    MOON_ILLUM = "Moon Illumination"



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# -----------------------
# Standard Python imports
# ----------------------

import asyncio
import logging


from datetime import datetime
from argparse import Namespace, ArgumentParser
from typing import Sequence

# -------------------
# Third party imports
# -------------------

import decouple

from lica.asyncio.cli import execute
from lica.typing import OptStr

# --------------
# local imports
# -------------

from . import __version__
from .utils import parser as prs
from .dbase import aux_dbase_load, aux_dbase_save
from .download import (
    download_ida_single, 
    download_ida_range, 
    ida_names_by_seq_or_range, 
    ida_names_by_location
)
from .timeseries import (
    to_ecsv_single,
    to_ecsv_range,
    to_ecsv_combine,
    NoCoordinatesError,
)

# ----------------
# Module constants
# ----------------

DESCRIPTION = "Full Download, Tranform and Combine IDA data from NextCloud Server"

# -----------------------
# Module global variables
# -----------------------

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


# ===========
# Generic API
# ===========


async def pipe_single(
    base_url: str,
    ida_base_dir: OptStr,
    ecsv_base_dir: OptStr,
    name: str,
    month: OptStr,
    exact: OptStr,
    fix: bool,
    timeout: int,
) -> None:
    await download_ida_single(base_url, ida_base_dir, name, month, exact, timeout)
    await asyncio.to_thread(
        to_ecsv_single, ida_base_dir, name, month, exact, ecsv_base_dir, fix
    )


async def pipe_range(
    base_url: str,
    ida_base_dir: OptStr,
    ecsv_base_dir: OptStr,
    name: str,
    since: datetime,
    until: datetime,
    skip_download: bool,
    oname: str,
    fix: bool,
    concurrent: int,
    timeout: int,
) -> None:
    if not skip_download:
        await download_ida_range(
            base_url, ida_base_dir, name, since, until, concurrent, timeout
        )
    await asyncio.to_thread(
        to_ecsv_range, ida_base_dir, name, ecsv_base_dir, since, until, fix
    )
    await asyncio.to_thread(to_ecsv_combine, ecsv_base_dir, name, since, until, oname)


async def pipe_photometers(
    base_url: str,
    ida_base_dir: OptStr,
    ecsv_base_dir: OptStr,
    rang: Sequence[int],
    seq: Sequence[int],
    since: datetime,
    until: datetime,
    skip_download: bool,
    oname: str,
    fix: bool,
    concurrent: int,
    timeout: int,
) -> None:
    names = ida_names_by_seq_or_range(seq, rang)
    if not skip_download:
        for name in names:
            await download_ida_range(base_url, ida_base_dir, name, since, until, concurrent, timeout)
    for name in names:
        await asyncio.to_thread(
            to_ecsv_range, ida_base_dir, name, ecsv_base_dir, since, until, fix
        )
        await asyncio.to_thread(
            to_ecsv_combine, ecsv_base_dir, name, since, until, oname
        )

async def pipe_location(
    base_url: str,
    ida_base_dir: OptStr,
    ecsv_base_dir: OptStr,
    lon: float,
    lat: float,
    radius: float,
    since: datetime,
    until: datetime,
    skip_download: bool,
    oname: str,
    fix: bool,
    concurrent: int,
    timeout: int,
) -> None:
    names = await ida_names_by_location(
        base_url, ida_base_dir, lon, lat, radius, timeout
    )
    if not skip_download:
        for name in names:
            await download_ida_range(base_url, ida_base_dir, name, since, until, concurrent, timeout)
    for name in names:
        await asyncio.to_thread(
            to_ecsv_range, ida_base_dir, name, ecsv_base_dir, since, until, fix
        )
        await asyncio.to_thread(
            to_ecsv_combine, ecsv_base_dir, name, since, until, oname
        )

# ================================
# COMMAND LINE INTERFACE FUNCTIONS
# ================================

async def cli_pipe_single(args: Namespace) -> None:
    await pipe_single(
        base_url=args.base_url,
        ida_base_dir=args.in_dir,
        ecsv_base_dir=args.out_dir,
        name=args.name,
        month=args.month,
        exact=args.exact,
        fix=True if args.fix else False,
        timeout=args.timeout,
    )


async def cli_pipe_range(args: Namespace) -> None:
    await pipe_range(
        base_url=args.base_url,
        ida_base_dir=args.in_dir,
        ecsv_base_dir=args.out_dir,
        name=args.name,
        since=args.since,
        until=args.until,
        skip_download=args.skip_download,
        oname=args.out_filename,
        fix=True if args.fix else False,
        concurrent=args.concurrent,
        timeout=args.timeout,
    )


async def cli_pipe_photometers(args: Namespace) -> None:
    await pipe_photometers(
        base_url=args.base_url,
        ida_base_dir=args.in_dir,
        ecsv_base_dir=args.out_dir,
        seq=args.list,
        rang=args.range,
        since=args.since,
        until=args.until,
        skip_download=args.skip_download,
        oname=args.out_filename,
        fix=True if args.fix else False,
        concurrent=args.concurrent,
        timeout=args.timeout,
    )


async def cli_pipe_location(args: Namespace) -> None:
    await pipe_location(
        base_url=args.base_url,
        ida_base_dir=args.in_dir,
        ecsv_base_dir=args.out_dir,
        lon=args.longitude,
        lat=args.latitude,
        radius=args.radius,
        since=args.since,
        until=args.until,
        skip_download=args.skip_download,
        oname=args.out_filename,
        fix=True if args.fix else False,
        concurrent=args.concurrent,
        timeout=args.timeout,
    )


def add_args(parser: ArgumentParser) -> ArgumentParser:
    # Now parse the application specific parts
    subparser = parser.add_subparsers(dest="command")
    parser_single = subparser.add_parser(
        "single",
        parents=[
            prs.name(),
            prs.out_dir("ECSV"),
            prs.inout_file("IDA", "combined ECSV", in_dir_exists=False),
            prs.mon_single(),
            prs.timeout(),
            prs.fix(),
        ],
        help="Process single monthly file from a photometer",
    )
    parser_single.set_defaults(func=cli_pipe_single)
    parser_range = subparser.add_parser(
        "range",
        parents=[
            prs.name(),
            prs.out_dir("ECSV"),
            prs.inout_file("IDA", "combined ECSV", in_dir_exists=False),
            prs.mon_range(),
            prs.concurrent(),
            prs.fix(),
            prs.skip(),
        ],
        help="Process a month range from a photometer",
    )
    parser_range.set_defaults(func=cli_pipe_range)
    parser_phots = subparser.add_parser(
        "photometers",
        parents=[
            prs.phot_range(),
            prs.out_dir("ECSV"),
            prs.inout_file("IDA", "combined ECSV", in_dir_exists=False),
            prs.mon_range(),
            prs.concurrent(),
            prs.fix(),
            prs.skip(),
        ],
        help="Download a month range for selected photometers",
    )
    parser_phots.set_defaults(func=cli_pipe_photometers)
    parser_location = subparser.add_parser(
        "near",
        parents=[
            prs.location(),
            prs.out_dir("ECSV"),
            prs.inout_file("IDA", "combined ECSV", in_dir_exists=False),
            prs.mon_range(),
            prs.concurrent(),
            prs.fix(),
            prs.skip(),
        ],
        help="Process a month range from photometers near a given location",
    )
    parser_location.set_defaults(func=cli_pipe_location)
    return parser


async def cli_pipeline(args: Namespace) -> None:
    """The main entry point specified by pyproject.toml"""
    args.base_url = decouple.config("IDA_URL")
    aux_dbase_load()
    try:
        await args.func(args)
    except NoCoordinatesError:
        pass
    aux_dbase_save()
    log.info("done!")


def main() -> None:
    execute(
        main_func=cli_pipeline,
        add_args_func=add_args,
        name="tess-ida-pipe",
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# ------------------------
# Standard Python imports
# -----------------------

import os
import hashlib

from datetime import datetime

# -------------------
# Third party imports
# -------------------

from dateutil.relativedelta import relativedelta
from lica.typing import OptStr

# -------------------
# Auxiliary functions
# -------------------


def month_range(from_month: datetime, to_month: datetime) -> str:
    month = from_month
    while month <= to_month:
        yield month.strftime("%Y-%m")
        month += relativedelta(months=1)


def name_month(ida_file_path: str) -> tuple:
    name = os.path.basename(os.path.dirname(ida_file_path))
    month = os.path.splitext(os.path.basename(ida_file_path))[0].split("_")[1]
    return name, month


def to_phot_dir(base_dir: OptStr, name: str) -> str:
    cwd = os.getcwd()
    base_dir = cwd if base_dir is None else base_dir
    base_dir = os.path.join(cwd, base_dir) if not os.path.isabs(base_dir) else base_dir
    full_dir_path = os.path.join(base_dir, name)
    return full_dir_path


def makedirs(base_dir: OptStr, name: str) -> str:
    full_dir_path = to_phot_dir(base_dir, name)
    os.makedirs(full_dir_path, exist_ok=True)
    return full_dir_path


def v_or_n(value: str) -> OptStr:
    """Value or None function"""
    value = value.strip()
    return None if value.lower() in ["none", "unknown", ""] else value


def hash_func(file_path: str) -> str:
    """Compute a hash from the image"""
    BLOCK_SIZE = 1048576  # 1MByte, the size of each read from the file
    # md5() was the fastest algorithm I've tried
    file_hash = hashlib.md5()
    with open(file_path, "rb") as f:
        block = f.read(BLOCK_SIZE)
        while len(block) > 0:
            file_hash.update(block)
            block = f.read(BLOCK_SIZE)
    return file_hash.hexdigest()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2021
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

from datetime import datetime
from argparse import ArgumentParser

# ---------------------
# Thrid-party libraries
# ---------------------

from dateutil.relativedelta import relativedelta
from lica.validators import vmonth, vdir, vfloat


# ------------------------
# Own modules and packages
# ------------------------


def cur_month() -> datetime:
    return datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def prev_month() -> datetime:
    month = datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    return month - relativedelta(months=1)


# -----------------
# Auxiliary parsers
# -----------------


def name() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument("-n", "--name", type=str, required=True, help="Photometer name")
    return parser


def fix() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument("-f", "--fix", action="store_true", help="Fix unknown location")
    return parser


def skip() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-sd", "--skip-download", action="store_true", help="Skip download step"
    )
    return parser


def timeout() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--timeout",
        type=int,
        default=300,
        help="HTTP timeout in seconds (defaults to %(default)s) sec.",
    )
    return parser


def inout_dirs(
    tag_in: str = "", tag_out: str = "", in_dir_exists: bool = True
) -> ArgumentParser:
    validator = vdir if in_dir_exists else str
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-i",
        "--in-dir",
        type=validator,
        default=None,
        help=f"Input {tag_in} base directory",
    )
    parser.add_argument(
        "-o",
        "--out-dir",
        type=str,
        default=None,
        help=f"Output {tag_out} base directory",
    )
    return parser


def out_dir(tag: str = "") -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-o", "--out-dir", type=str, default=None, help=f"Output {tag} base directory"
    )
    return parser


def inout_file(
    tag_in: str = "", tag_out: str = "", in_dir_exists: bool = True
) -> ArgumentParser:
    validator = vdir if in_dir_exists else str
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-i",
        "--in-dir",
        type=validator,
        default=None,
        help=f"Input {tag_in} base directory",
    )
    parser.add_argument(
        "-on",
        "--out-filename",
        type=str,
        default=None,
        help=f"Optional output {tag_out} file name",
    )
    return parser


def concurrent() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-c",
        "--concurrent",
        type=int,
        metavar="<N>",
        choices=[1, 2, 4, 6, 8],
        default=4,
        help="Number of concurrent downloads (defaults to %(default)s)",
    )
    parser.add_argument(
        "--timeout",
        type=int,
        default=300,
        help="HTTP timeout in seconds (defaults to %(default)s sec.)",
    )
    return parser


def mon_single() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-e", "--exact", type=str, default=None, help="Specific monthly file name"
    )
    group.add_argument(
        "-m",
        "--month",
        type=vmonth,
        default=None,
        metavar="<YYYY-MM>",
        help="Year and Month",
    )
    return parser


def mon_range() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-s",
        "--since",
        type=vmonth,
        default=prev_month(),
        metavar="<YYYY-MM>",
        help="Year and Month (defaults to %(default)s)",
    )
    parser.add_argument(
        "-u",
        "--until",
        type=vmonth,
        default=cur_month(),
        metavar="<YYYY-MM>",
        help="Year and Month (defaults to %(default)s)",
    )
    return parser


def phot_range() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-l",
        "--list",
        type=int,
        default=None,
        nargs="+",
        metavar="<N>",
        help="Photometer number list",
    )
    group.add_argument(
        "-r",
        "--range",
        type=int,
        default=None,
        metavar="<N>",
        nargs=2,
        help="Photometer number range",
    )
    return parser


def location() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-lo",
        "--longitude",
        type=vfloat,
        required=True,
        metavar="<LON>",
        help="Longitude (decimal degrees)",
    )
    parser.add_argument(
        "-la",
        "--latitude",
        type=vfloat,
        required=True,
        metavar="<LAT>",
        help="Latitude (decimal degrees)",
    )
    parser.add_argument(
        "-ra",
        "--radius",
        type=vfloat,
        default=10,
        metavar="<R>",
        help="Search radius (Km) (defaults to %(default)s Km)",
    )
    return parser



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# -----------------------
# Standard Python imports
# ----------------------

import os
import glob
import logging

from datetime import datetime
from argparse import Namespace, ArgumentParser
from typing import Union, Dict, Any

# -------------------
# Third party imports
# -------------------

import numpy as np

import astropy.units as u
from astropy.table import vstack
from astropy.timeseries import TimeSeries
from astropy.coordinates import EarthLocation
from astroplan import Observer

from lica.cli import execute
from lica.typing import OptStr

# --------------
# local imports
# -------------

from . import __version__
from .dbase import (
    aux_dbase_load,
    aux_dbase_save,
    aux_table_hashes_lookup,
    aux_table_hashes_insert,
    aux_table_hashes_update,
    aux_table_coords_lookup,
)

from .constants import (
    TESSW_COLS as TEW,
    TESS4C_COLS as T4C,
    TIMESERIES_COLS as TS,
    IDA_KEYWORDS as IKW,
    IDA_HEADER_LEN,
)
from .utils.utils import (
    to_phot_dir,
    makedirs,
    v_or_n,
    month_range,
    name_month,
    hash_func,
)

from .utils import parser as prs

# ----------------
# Module constants
# ----------------

DESCRIPTION = "Transform TESS-W IDA monthly files to ECSV"
OptDate = Union[datetime, None]

# Column names
IDA_NAMES = TEW.values()
IDA_NAMES_4C = T4C.values()

# data types for column names
IDA_DTYPES = {
    TEW.UTC_TIME: str,
    TEW.LOCAL_TIME: str,
    TEW.BOX_TEMP: float,
    TEW.SKY_TEMP: float,
    TEW.FREQ: float,
    TEW.MAG: float,
    TEW.ZP: float,
    TEW.SEQ_NUM: int,
}
IDA_DTYPES_4C = {
    T4C.UTC_TIME: str,
    T4C.LOCAL_TIME: str,
    T4C.BOX_TEMP: float,
    T4C.SKY_TEMP: float,
    T4C.FREQ1: float,
    T4C.MAG1: float,
    T4C.ZP1: float,
    T4C.FREQ2: float,
    T4C.MAG2: float,
    T4C.ZP2: float,
    T4C.FREQ3: float,
    T4C.MAG3: float,
    T4C.ZP3: float,
    T4C.FREQ4: float,
    T4C.MAG4: float,
    T4C.ZP4: float,
    "Sequence Number": int,
}

# Exclude these columns from the final Table
IDA_EXCLUDE = (TEW.LOCAL_TIME,)


# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# -----------------
# Custom Exceptions
# -----------------


class NoCoordinatesError(Exception):
    pass


# -------------------
# Auxiliary functions
# -------------------

# =============
# Work Routines
# =============


def ida_metadata(path: str, fix: bool) -> Dict[str, Any]:
    # Reads the whole header, strips off the starting '# ' and trailing '\n'
    name, month = name_month(path)
    with open(path) as f:
        lines = [next(f)[2:-1] for _ in range(IDA_HEADER_LEN)]
    lines = lines[:-13]  # Strips off the last 13 lines (including comments)
    # make  key-value pairs
    pairs = [line.split(": ") for line in lines]
    pairs = list(filter(lambda x: len(x) == 2, pairs))
    pairs[2][0] = str(IKW.LICENSE)  # patch it, real keyword is too long ...
    # Make a dict out of (keyword, value) pairs
    header = dict(pairs)
    # Convert values from strings to numeric values or nested dictonaries
    header[IKW.NUM_HEADERS] = int(header[IKW.NUM_HEADERS])
    header[IKW.NUM_CHANNELS] = int(header[IKW.NUM_CHANNELS])
    observer, affil = header[IKW.OBSERVER].split("/")
    header[IKW.OBSERVER] = {"observer": v_or_n(observer), "affiliation": v_or_n(affil)}
    place, town, sub_region, region, country = header[IKW.LOCATION].split("/")
    header[IKW.LOCATION] = {
        "place": v_or_n(place),
        "town": v_or_n(town),
        "sub_region": v_or_n(sub_region),
        "region": v_or_n(region),
        "country": v_or_n(country),
    }
    latitude, longitude, height = header[IKW.POSITION].split(",")
    try:
        header[IKW.POSITION] = {
            "latitude": float(v_or_n(latitude)),
            "longitude": float(v_or_n(longitude)),
            "height": float(v_or_n(height)) if v_or_n(height) else 0.0,
        }
    except TypeError:
        if not fix:
            log.error("[%s] [%s] Unknown Observer coordinates", name, month)
            log.error(
                "[%s] [%s] Add coordinates to aux. coordinates table & re-run with --fix",
                name,
                month,
            )
            raise NoCoordinatesError
        coords = next(aux_table_coords_lookup(name))
        if not coords:
            log.error(
                "[%s] [%s] Could not find alternative coordinates in the aux. coordinates table",
                name,
                month,
            )
            raise NoCoordinatesError
        _, lati, longi, h = coords
        log.warning(
            "[%s] [%s] Fixed alternative coordinates (lat: %f, long: %f, h: %f) from the adm coordinates table",
            name,
            month,
            lati,
            longi,
            h,
        )
        header[IKW.POSITION] = {"latitude": lati, "longitude": longi, "height": h}
    header[IKW.FOV] = float(header[IKW.FOV])
    header[IKW.COVER_OFFSET] = float(header[IKW.COVER_OFFSET])
    header[IKW.NUM_COLS] = int(header[IKW.NUM_COLS])
    if header[IKW.NUM_CHANNELS] == 1:
        assert header[IKW.NUM_COLS] == 8
        header[IKW.ZP] = float(
            header[IKW.ZP].split("(")[0]
        )  # get rid of possible comments on the right of (
        az, zen = header[IKW.AIM][1:-1].split(",")
        header[IKW.AIM] = {"azimuth": float(az), "zenital": float(zen)}
    else:
        assert header[IKW.NUM_CHANNELS] == 4
        assert header[IKW.NUM_COLS] == 17
        header["Filters per channel"] = [
            f.strip()[1:-1] for f in header["Filters per channel"][1:-1].split(",")
        ]
        header[IKW.ZP] = [float(zp) for zp in header[IKW.ZP][1:-1].split(",")]
        az1, zen1, az2, zen2, az3, zen3, az4, zen4 = header[IKW.AIM][1:-1].split(",")
        header[IKW.AIM] = [
            {"azimuth": float(az1), "zenital": float(zen1)},
            {"azimuth": float(az2), "zenital": float(zen2)},
            {"azimuth": float(az3), "zenital": float(zen3)},
            {"azimuth": float(az4), "zenital": float(zen4)},
        ]
    return header


def ida_to_table(path: str, fix: bool) -> TimeSeries:
    header = ida_metadata(path, fix)
    nchannels = header[IKW.NUM_CHANNELS]
    names = IDA_NAMES if nchannels == 1 else IDA_NAMES_4C
    converters = IDA_DTYPES if nchannels == 1 else IDA_DTYPES_4C
    table = TimeSeries.read(
        path,
        time_column=IDA_NAMES[0],
        format="ascii.basic",
        delimiter=";",
        data_start=0,
        names=names,
        exclude_names=IDA_EXCLUDE,
        converters=converters,
        guess=False,
    )
    table.meta["ida"] = header
    del table.meta["comments"]
    # Convert to quiatities by adding units
    table[TEW.BOX_TEMP] = table[TEW.BOX_TEMP] * u.deg_C
    table[TEW.SKY_TEMP] = table[TEW.SKY_TEMP] * u.deg_C
    if nchannels == 1:
        table[TEW.FREQ] = table[TEW.FREQ] * u.Hz
        table[TEW.MAG] = table[TEW.MAG] * u.mag() / u.arcsec**2
    else:
        table[T4C.FREQ1] = table[T4C.FREQ1] * u.Hz
        table[T4C.MAG1] = table[T4C.MAG1] * u.mag() / u.arcsec**2
        table[T4C.FREQ2] = table[T4C.FREQ2] * u.Hz
        table[T4C.MAG2] = table[T4C.MAG2] * u.mag() / u.arcsec**2
        table[T4C.FREQ3] = table[T4C.FREQ3] * u.Hz
        table[T4C.MAG3] = table[T4C.MAG3] * u.mag() / u.arcsec**2
        table[T4C.FREQ4] = table[T4C.FREQ4] * u.Hz
        table[T4C.MAG4] = table[T4C.MAG4] * u.mag() / u.arcsec**2
    return table


def add_columns(table: TimeSeries, name: str, month: str) -> None:
    latitude = table.meta["ida"][IKW.POSITION]["latitude"]
    longitude = table.meta["ida"][IKW.POSITION]["longitude"]
    height = table.meta["ida"][IKW.POSITION]["height"]
    obs_name = table.meta["ida"][IKW.OBSERVER]["observer"]
    zenital = table.meta["ida"][IKW.AIM]["zenital"]  # Only valid for TESS-W
    location = EarthLocation(lat=latitude, lon=longitude, height=height)
    observer = Observer(name=obs_name, location=location)
    log.info("[%s] [%s] Adding new %s column", name, month, TS.SUN_ALT)
    sun_altaz = observer.sun_altaz(table["time"])
    table[TS.SUN_ALT] = np.round(sun_altaz.alt.deg, 2) * u.deg
    if zenital != 0.0:
        log.info("[%s] [%s] Adding new %s column", name, month, TS.SUN_AZ)
        table[TS.SUN_AZ] = np.round(sun_altaz.az.deg, 2) * u.deg
    log.info("[%s] [%s] Adding new %s column", name, month, TS.MOON_ALT)
    moon_altaz = observer.moon_altaz(table["time"])
    table[TS.MOON_ALT] = np.round(moon_altaz.alt.deg) * u.deg
    if zenital != 0.0:
        log.info("[%s] [%s] Adding new %s column", name, month, TS.MOON_AZ)
        table[TS.MOON_AZ] = np.round(moon_altaz.az.deg, 2) * u.deg
    log.info("[%s] [%s] Adding new %s column", name, month, TS.MOON_ILLUM)
    table[TS.MOON_ILLUM] = np.round(observer.moon_illumination(table["time"]), 3)


def create_table(path: str, fix: bool) -> TimeSeries:
    """Create TimeSeries or loads table from IDA file"""
    name, month = name_month(path)
    log.info("[%s] [%s] Creating a Time Series from IDA file: %s", name, month, path)
    table = ida_to_table(path, fix)
    add_columns(table, name, month)
    return table


def load_table(path: str) -> TimeSeries:
    """Read TimeSeries table from ECSV file"""
    return TimeSeries.read(path, format="ascii.ecsv", delimiter=",")


def save_table(table: TimeSeries, path: str) -> None:
    """Read TimeSeries table from ECSV file"""
    name, month = name_month(path)
    log.info("[%s] [%s] Saving Time Series to ECSV file: %s", name, month, path)
    table.write(
        path, format="ascii.ecsv", delimiter=",", fast_writer=True, overwrite=True
    )


def append_table(acc: TimeSeries, table: TimeSeries) -> TimeSeries:
    return vstack([acc, table])


def do_to_ecsv_single(in_path: str, out_path: str, fix: bool) -> None:
    name, month = name_month(in_path)
    data = [os.path.basename(in_path), hash_func(in_path)]
    result = next(aux_table_hashes_lookup(data[0]))
    if result:
        _, stored_hash_str = result
        if data[1] != stored_hash_str or not os.path.isfile(out_path):
            aux_table_hashes_update(data)
            table = create_table(in_path, fix)
            save_table(table, out_path)
        else:
            log.info(
                "[%s] [%s] Time Series already in ECSV file: %s", name, month, out_path
            )
    else:
        aux_table_hashes_insert(data)
        table = create_table(in_path, fix)
        save_table(table, out_path)


# ===========
# Generic API
# ===========


def to_ecsv_single(
    base_dir: OptStr, name: str, month: OptDate, exact: OptStr, out_dir: str, fix: bool
) -> None:
    in_dir_path = to_phot_dir(base_dir, name)
    filename = name + "_" + month.strftime("%Y-%m") + ".dat" if not exact else exact
    in_path = os.path.join(in_dir_path, filename)
    out_dir_path = makedirs(out_dir, name)
    filename = os.path.splitext(filename)[0]
    out_path = os.path.join(out_dir_path, filename + ".ecsv")
    do_to_ecsv_single(in_path, out_path, fix)


def to_ecsv_range(
    base_dir: OptStr,
    name: str,
    out_dir: str,
    since: datetime,
    until: datetime,
    fix: bool,
) -> None:
    in_dir_path = to_phot_dir(base_dir, name)
    months = [m for m in month_range(since, until)]
    search_path = os.path.join(in_dir_path, "*.dat")
    candidate_path = list()
    for path in sorted(glob.iglob(search_path)):
        candidate_month = os.path.splitext(os.path.basename(path))[0].split("_")[1]
        if candidate_month in months:
            candidate_path.append(path)
    for in_path in candidate_path:
        filename = os.path.splitext(os.path.basename(in_path))[0] + ".ecsv"
        dirname = makedirs(out_dir, name)
        out_path = os.path.join(dirname, filename)
        do_to_ecsv_single(in_path, out_path, fix)


def to_ecsv_combine(
    base_dir: OptStr, name: str, since: datetime, until: datetime, oname: str
) -> None:
    in_dir_path = to_phot_dir(base_dir, name)
    months = [m for m in month_range(since, until)]
    log.info(
        "[%s] Combining %d months from %s to %s into a single ECSV",
        name,
        len(months),
        months[0],
        months[-1],
    )
    search_path = os.path.join(in_dir_path, "stars*.ecsv")
    candidate_path = list()
    for path in sorted(glob.iglob(search_path)):
        candidate_month = os.path.splitext(os.path.basename(path))[0].split("_")[1]
        if candidate_month in months:
            candidate_path.append(path)
    if len(candidate_path) < 1:
        log.warning("[%s] No tables to combine. Check range input parameters.", name)
        return
    acc_table = load_table(candidate_path[0])
    acc_table.meta["combined"] = [os.path.basename(candidate_path[0])]
    for in_path in candidate_path[1:]:
        table = load_table(in_path)
        acc_table = append_table(acc_table, table)
        acc_table.meta["combined"].append(os.path.basename(in_path))
    dirname = os.path.dirname(candidate_path[0])
    filename = (
        f"{name}_{since.strftime('%Y%m')}-{until.strftime('%Y%m')}.ecsv"
        if oname is None
        else oname
    )
    path = os.path.join(dirname, filename)
    log.info("[%s] Saving combined Time Series to ECSV file: %s", name, path)
    acc_table.write(
        path, format="ascii.ecsv", delimiter=",", fast_writer=True, overwrite=True
    )


# ================================
# COMMAND LINE INTERFACE FUNCTIONS
# ================================


def cli_to_ecsv_single(args: Namespace) -> None:
    to_ecsv_single(
        base_dir=args.in_dir,
        name=args.name,
        month=args.month,
        exact=args.exact,
        out_dir=args.out_dir,
        fix=True if args.fix else False,
    )


def cli_to_ecsv_range(args: Namespace) -> None:
    to_ecsv_range(
        base_dir=args.in_dir,
        name=args.name,
        out_dir=args.out_dir,
        since=args.since,
        until=args.until,
        fix=True if args.fix else False,
    )


def cli_to_ecsv_combine(args: Namespace) -> None:
    to_ecsv_combine(
        base_dir=args.in_dir,
        name=args.name,
        since=args.since,
        until=args.until,
        oname=args.out_filename,
    )


def add_args(parser: ArgumentParser) -> None:
    # Now parse the application specific parts
    subparser = parser.add_subparsers(dest="command")
    parser_single = subparser.add_parser(
        "single",
        parents=[
            prs.name(),
            prs.mon_single(),
            prs.inout_dirs("IDA", "ECSV"),
            prs.fix(),
        ],
        help="Convert to ECSV a single monthly file from a photometer",
    )
    parser_single.set_defaults(func=cli_to_ecsv_single)

    parser_range = subparser.add_parser(
        "range",
        parents=[prs.name(), prs.mon_range(), prs.inout_dirs("IDA", "ECSV"), prs.fix()],
        help="Convert to ECSV a range of IDA monthly files from a photometer",
    )
    parser_range.set_defaults(func=cli_to_ecsv_range)

    parser_comb = subparser.add_parser(
        "combine",
        parents=[prs.name(), prs.mon_range(), prs.inout_file("ECSV", "combined")],
        help="Combines a range of monthly ECSV files into a single ECSV",
    )
    parser_comb.set_defaults(func=cli_to_ecsv_combine)


def cli_to_ecsv(args: Namespace) -> None:
    """The main entry point specified by pyprojectable.toml"""
    aux_dbase_load()
    try:
        args.func(args)
    except NoCoordinatesError:
        pass
    aux_dbase_save()
    log.info("done!")


def main() -> None:
    execute(
        main_func=cli_to_ecsv,
        add_args_func=add_args,
        name="tess-ida-ecsv",
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



"""
This test module only test the Command Line Interface, with its various options.
The command output is written to a separate log file so that stdout is clean for unittest.

From the project base dir dir, run as:

    python -m unittest -v test.test_cli.TestDownload
    python -m unittest -v test.test_cli.TestECSV
    python -m unittest -v test.test_cli.TestPipeline
    <etc>

or the complete suite:

    python -m unittest -v test.test_cli

"""

import unittest
from shlex import split
from subprocess import run

import decouple

class TestDownload(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.log = f"{cls.__name__}.log"
        run(split(f"touch {cls.log}"))
        run(split(f"truncate --size 0 {cls.log}"))

    def test_1_single(self):
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} single -n stars289 -m 2023-06 -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_2_exact(self):
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} single -n stars201 -e stars201_2020-02_-1.dat -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} single -n stars201 -e stars201_2020-02_61.dat -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_3_range(self):
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} range -n stars289 -s 2023-06 -u 2023-09 -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_4_phot_list(self):
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} photometers --list 1 5 33 44 85 -s 2022-01 -u 2022-05 -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_5_phot_range(self):
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} photometers --range 300 305 -s 2022-01 -u 2022-05 -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_6_phot_near(self):
        result = run(
            split(
                f"tess-ida-get --log-file {self.log} near -lo -3.703790 -la 40.416775 -ra 50 -o IDA"
            )
        )
        self.assertEqual(result.returncode, 0)


class TestECSV(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.log = f"{cls.__name__}.log"
        run(split(f"touch {cls.log}"))
        run(split(f"truncate --size 0 {cls.log}"))
        run(split("rm -fr ECSV"))

    def test_1_single(self):
        result = run(
            split(
                f"tess-ida-ecsv --log-file {self.log} single -n stars289 -m 2023-06 -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_2_exact(self):
        result = run(
            split(
                f"tess-ida-ecsv --log-file {self.log} single -n stars201 -e stars201_2020-02_-1.dat -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-ecsv --log-file {self.log} single -n stars201 -e stars201_2020-02_61.dat -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_3_range(self):
        result = run(
            split(
                f"tess-ida-ecsv --log-file {self.log} range -n stars289 -s 2023-06 -u 2023-09 -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_4_combine(self):
        result = run(
            split(
                f"tess-ida-ecsv --log-file {self.log} combine -n stars289 -s 2023-06 -u 2023-09 -i IDA -on out.ecsv"
            )
        )
        self.assertEqual(result.returncode, 0)


class TestPipeline(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.log = f"{cls.__name__}.log"
        run(split(f"touch {cls.log}"))
        run(split(f"truncate --size 0 {cls.log}"))
        run(split("rm -fr IDA ECSV"))

    def test_1_single(self):
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} single -n stars289 -m 2023-06 -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_2_exact(self):
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} single -n stars201 -e stars201_2020-02_-1.dat -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} single -n stars201 -e stars201_2020-02_61.dat -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_3_range(self):
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} range -n stars289 -s 2023-06 -u 2023-09 -i IDA -o ECSV -on output.ecsv"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_4_photometers_range(self):
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} photometers -r 1 5  2017-06 -u 2017-09 -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_5_photometers_list(self):
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} photometers -l 1 5 -s 2017-06 -u 2017-09 -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_6_photometers_near(self):
        result = run(
            split(
                f"tess-ida-pipe --log-file {self.log} near -lo -3.703790 -la 40.416775 -ra 50 -s 2017-06 -u 2017-09 -i IDA -o ECSV"
            )
        )
        self.assertEqual(result.returncode, 0)


class TestDBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.log = f"{cls.__name__}.log"
        run(split(f"rm -fr {decouple.config('DATABASE_FILE')}"))

    def test_1_schema_create(self):
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose schema create"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_2_coords_add(self):
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords add -n stars200 -lo -3.5 -la 40.5 -he 650"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords add -n stars201 -lo -2 -la 41.2 -he 300"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_3_coords_list(self):
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords list -n stars200"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords list"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_4_coords_update(self):
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords update -n stars200 -lo 3.5 -la 40.8 -he 0"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords list"
            )
        )
        self.assertEqual(result.returncode, 0)

    def test_5_coords_delete(self):
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords delete -n stars200"
            )
        )
        self.assertEqual(result.returncode, 0)
        result = run(
            split(
                f"tess-ida-db --log-file {self.log} --verbose coords list"
            )
        )
        self.assertEqual(result.returncode, 0)



if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# -----------------------
# Standard Python imports
# ----------------------

import os
import io
import csv
import math
import asyncio
import logging
import functools

from datetime import datetime
from argparse import Namespace, ArgumentParser
from typing import Dict, Tuple, Sequence, Optional, Any

# -------------------
# Third party imports
# -------------------

import decouple
import aiohttp
import aiofiles


from lica.asyncio.cli import execute
from lica.misc import group
from lica.typing import OptStr

# --------------
# local imports
# -------------

from . import __version__
from .utils import parser as prs
from .utils.utils import month_range, makedirs, name_month

# ----------------
# Module constants
# ----------------

DESCRIPTION = "Get TESS-W IDA monthly files from NextCloud server"

# -----------------------
# Module global variables
# -----------------------

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


def distance(
    coords_A: Tuple[float, float], coords_B: Tuple[float, float]
) -> Optional[float]:
    """
    Compute approximate geographical distance (arc) [meters] between two points on Earth
    Coods_A and Coords_B are tuples (longitude, latitude)
    Accurate for small distances only
    """
    EARTH_RADIUS = 6371009.0  # in meters

    long_A = coords_A[0]
    long_B = coords_B[0]
    lat_A = coords_A[1]
    lat_B = coords_B[1]
    try:
        delta_long = math.radians(long_A - long_B)
        delta_lat = math.radians(lat_A - lat_B)
        mean_lat = math.radians((lat_A + lat_B) / 2)
        result = round(
            EARTH_RADIUS
            * math.sqrt(delta_lat**2 + (math.cos(mean_lat) * delta_long) ** 2),
            0,
        )
    except TypeError:
        log.error("Algo malo pasṕ")
        result = None
    return result


def filter_by_distance(
    longitude: float, latitude: float, radius: float, item: Dict[str, Sequence[int]]
) -> bool:
    d = distance((item["longitude"], item["latitude"]), (longitude, latitude))
    return True if d <= radius else False


def float_coords(item: Dict[str, Any]) -> Dict[str, Any]:
    """
    Converts coordinates in items read by CSV reader into floating values
    """
    item["longitude"] = float(item["longitude"])
    item["latitude"] = float(item["latitude"])
    return item


async def do_get_location_list(
    base_url: str, ida_base_dir: str, timeout: int
) -> Sequence:
    target_file = "geolist.csv"
    url = os.path.join(base_url, target_file)
    result = []
    timeout = aiohttp.ClientTimeout(total=timeout)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        async with session.get(url, params=None) as resp:
            if resp.status == 404:
                log.warn("No such file exits: %s", target_file)
                return result
            log.info("[%s] GET %s [%d OK]", target_file, resp.url, resp.status)
            contents = await resp.text()
            if len(contents) == 0:
                log.warning("File size is 0. URL scheme may have changed ...")
            with io.StringIO(contents) as fd:
                # reader = csv.reader(fd, delimiter=';')
                reader = csv.DictReader(fd, delimiter=";")
                result = list(map(float_coords, reader))
    return result


async def do_ida_single(
    session, base_url: str, ida_base_dir: str, name: str, month: OptStr, exact: OptStr
) -> None:
    target_file = name + "_" + month + ".dat" if not exact else exact
    url = os.path.join(base_url, name, target_file)
    _, month1 = name_month(target_file)
    async with session.get(url, params=None) as resp:
        if resp.status == 404:
            log.warn("[%s] No monthly file exits: %s", name, target_file)
            return
        log.info("[%s] [%s] GET %s [%d OK]", name, month1, resp.url, resp.status)
        contents = await resp.text()
    if len(contents) == 0:
        log.warning("File size is 0. URL scheme may have changed ...")
    full_dir_path = await asyncio.to_thread(makedirs, ida_base_dir, name)
    file_path = os.path.join(full_dir_path, target_file)
    async with aiofiles.open(file_path, mode="w") as f:
        log.info("[%s] [%s] Writing %s", name, month1, file_path)
        await f.write(contents)


async def do_ida_range(
    session,
    base_url: str,
    ida_base_dir: str,
    name: str,
    since: datetime,
    until: datetime,
    N: int,
) -> None:
    for grp in group(N, month_range(since, until)):
        tasks = [
            asyncio.create_task(
                do_ida_single(session, base_url, ida_base_dir, name, m, None)
            )
            for m in grp
        ]
        await asyncio.gather(*tasks)


# ===========
# Generic API
# ===========


async def ida_names_by_location(
    base_url: str,
    ida_base_dir: str,
    lon: float,
    lat: float,
    radius: float,
    timeout: int,
) -> Tuple[str]:
    by_distance = functools.partial(filter_by_distance, lon, lat, 1000 * radius)
    result = await do_get_location_list(base_url, ida_base_dir, timeout)
    return tuple(item["name"] for item in filter(by_distance, result))


def ida_names_by_seq_or_range(seq: Sequence[int], rang: Sequence[int]) -> Tuple[str]:
    if seq is not None:
        result = tuple("stars" + str(i) for i in sorted(seq))
    else:
        rang = sorted(rang)
        result = tuple("stars" + str(i) for i in range(rang[0], rang[1] + 1))
    return result


async def download_ida_single(
    base_url: str,
    ida_base_dir: str,
    name: str,
    month: OptStr,
    exact: OptStr,
    timeout: int,
) -> None:
    timeout = aiohttp.ClientTimeout(total=timeout)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        if not exact:
            month = month.strftime("%Y-%m")
        await do_ida_single(session, base_url, ida_base_dir, name, month, exact)


async def download_ida_range(
    base_url: str,
    ida_base_dir: str,
    name: str,
    since: datetime,
    until: datetime,
    concurrent: int,
    timeout: int,
) -> None:
    timeout = aiohttp.ClientTimeout(total=timeout)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        await do_ida_range(
            session, base_url, ida_base_dir, name, since, until, concurrent
        )


async def ida_photometers(
    base_url: str,
    ida_base_dir: str,
    rang: Sequence[int],
    seq: Sequence[int],
    since: datetime,
    until: datetime,
    concurrent: int,
    timeout: int,
) -> None:
    for name in ida_names_by_seq_or_range(seq, rang):
        await download_ida_range(
            base_url, ida_base_dir, name, since, until, concurrent, timeout
        )


async def ida_location(
    base_url: str,
    ida_base_dir: str,
    lon: float,
    lat: float,
    radius: float,
    since: datetime,
    until: datetime,
    concurrent: int,
    timeout: int,
) -> None:
    names = await ida_names_by_location(
        base_url, ida_base_dir, lon, lat, radius, timeout
    )
    for name in names:
        await download_ida_range(
            base_url, ida_base_dir, name, since, until, concurrent, timeout
        )


# ================================
# COMMAND LINE INTERFACE FUNCTIONS
# ================================


async def cli_ida_single(args: Namespace) -> None:
    await download_ida_single(
        base_url=args.base_url,
        ida_base_dir=args.out_dir,
        name=args.name,
        month=args.month,
        exact=args.exact,
        timeout=args.timeout,
    )


async def cli_ida_range(args: Namespace) -> None:
    await download_ida_range(
        base_url=args.base_url,
        ida_base_dir=args.out_dir,
        name=args.name,
        since=args.since,
        until=args.until,
        concurrent=args.concurrent,
        timeout=args.timeout,
    )


async def cli_ida_photometers(args: Namespace) -> None:
    await ida_photometers(
        base_url=args.base_url,
        ida_base_dir=args.out_dir,
        seq=args.list,
        rang=args.range,
        since=args.since,
        until=args.until,
        concurrent=args.concurrent,
        timeout=args.timeout,
    )


async def cli_ida_location(args: Namespace) -> None:
    await ida_location(
        base_url=args.base_url,
        ida_base_dir=args.out_dir,
        lon=args.longitude,
        lat=args.latitude,
        radius=args.radius,
        since=args.since,
        until=args.until,
        concurrent=args.concurrent,
        timeout=args.timeout,
    )


def add_args(parser: ArgumentParser) -> ArgumentParser:
    # Now parse the application specific parts
    subparser = parser.add_subparsers(dest="command")
    parser_single = subparser.add_parser(
        "single",
        parents=[prs.name(), prs.out_dir("IDA"), prs.mon_single(), prs.timeout()],
        help="Download single monthly file from a photometer",
    )
    parser_single.set_defaults(func=cli_ida_single)
    parser_range = subparser.add_parser(
        "range",
        parents=[prs.name(), prs.out_dir("IDA"), prs.mon_range(), prs.concurrent()],
        help="Download a month range from a photometer",
    )
    parser_range.set_defaults(func=cli_ida_range)
    parser_phots = subparser.add_parser(
        "photometers",
        parents=[
            prs.phot_range(),
            prs.out_dir("IDA"),
            prs.mon_range(),
            prs.concurrent(),
        ],
        help="Download a month range for selected photometers",
    )
    parser_phots.set_defaults(func=cli_ida_photometers)
    parser_location = subparser.add_parser(
        "near",
        parents=[prs.location(), prs.out_dir("IDA"), prs.mon_range(), prs.concurrent()],
        help="Download a month range from photometers near a given location",
    )
    parser_location.set_defaults(func=cli_ida_location)
    return parser


async def cli_get_ida(args: Namespace) -> None:
    """The main entry point specified by pyproject.toml"""
    args.base_url = decouple.config("IDA_URL")
    await args.func(args)
    log.info("done!")


def main() -> None:
    execute(
        main_func=cli_get_ida,
        add_args_func=add_args,
        name="tess-ida-get",
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# -----------------------
# Standard Python imports
# ----------------------

import logging
import sqlite3

from typing import Union, Iterator
from collections.abc import Sequence


# -------------------
# Third party imports
# -------------------

import decouple


# --------------
# local imports
# -------------


# ----------------
# Module constants
# ----------------

DESCRIPTION = "Database utility to speed up pipeline processing"

OptRow = Union[tuple, None]

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-2])

try:
    theDatabaseFile = decouple.config("DATABASE_FILE")

    def aux_dbase_load() -> None:
        global theDatabaseFile
        log.info("Opening auxiliar database from %s", theDatabaseFile)
        _ = sqlite3.connect(theDatabaseFile)

    def aux_dbase_save() -> None:
        pass

    def aux_table_hashes_insert(data: Sequence[str, str]) -> None:
        global theDatabaseFile
        with sqlite3.connect(theDatabaseFile) as conn:
            conn.execute("INSERT INTO ecsv_t(filename, hash) VALUES(?,?)", data)
        conn.close()

    def aux_table_hashes_update(data: Sequence[str, str]) -> None:
        global theDatabaseFile
        with sqlite3.connect(theDatabaseFile) as conn:
            conn.execute(
                "UPDATE ecsv_t SET hash = ? WHERE filename = ?", (data[1], data[0])
            )
        conn.close()

    def aux_table_hashes_lookup(filename: str) -> Iterator[OptRow]:
        global theDatabaseFile
        with sqlite3.connect(theDatabaseFile) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "SELECT filename, hash FROM ecsv_t WHERE filename = ?", (filename,)
            )
        yield cursor.fetchone()
        conn.close()

    def aux_table_coords_lookup(name: str) -> Iterator[OptRow]:
        global theDatabaseFile
        with sqlite3.connect(theDatabaseFile) as conn:
            cursor = conn.cursor()
        cursor.execute(
            "SELECT phot_name, latitude, longitude, height FROM coords_t WHERE phot_name = ?",
            (name,),
        )
        yield cursor.fetchone()
        conn.close()

except decouple.UndefinedValueError:

    def aux_dbase_load() -> None:
        log.warning(
            "No Auxiliar database was configured. Check 'DATABASE_FILE' environment variable"
        )

    def aux_dbase_save() -> None:
        log.warning(
            "No Auxiliar database was configured. Check 'DATABASE_FILE' environment variable"
        )

    def aux_table_hashes_lookup(filename: str) -> Iterator[OptRow]:
        yield None

    def aux_table_hashes_insert(data: Sequence[str, str]) -> None:
        pass

    def aux_table_hashes_update(data: Sequence[str, str]) -> None:
        pass

    def aux_table_coords_lookup(name) -> Iterator[OptRow]:
        yield None



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# -----------------------
# Standard Python imports
# ----------------------

import os
import sys
import logging
import sqlite3

from importlib.resources import files
from argparse import Namespace, ArgumentParser

# -------------------
# Third party imports
# -------------------

import decouple

from lica.cli import execute
from lica.tabulate import paging


# --------------
# local imports
# -------------

from .. import __version__

# ----------------
# Module constants
# ----------------

DESCRIPTION = "Database utility to speed up pipeline processing"

_SQL_PKG = "tess.ida.dbase.sql"
_SQL_RES = "schema.sql"
SCHEMA_SQL_TEXT = files(_SQL_PKG).joinpath(_SQL_RES).read_text()

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# ================================
# COMMAND LINE INTERFACE FUNCTIONS
# ================================


def cli_schema_create(args: Namespace) -> None:
    dbase_path = decouple.config("DATABASE_FILE")
    if os.path.isfile(dbase_path):
        log.info("Deleting auxiliar database: %s", dbase_path)
        os.remove(dbase_path)
    output_dir = os.path.dirname(dbase_path)
    output_dir = os.getcwd() if not output_dir else output_dir
    os.makedirs(output_dir, exist_ok=True)
    log.info("Creating SQLite Schema on %s", dbase_path)
    try:
        with sqlite3.connect(dbase_path) as connection:
            connection.executescript(SCHEMA_SQL_TEXT)
    except sqlite3.OperationalError as e:
        log.error("Error creating the database: %s", e)
    finally:
        connection.close()


def cli_coords_add(args: Namespace) -> None:
    data = (args.name, args.latitude, args.longitude, args.height)
    dbase_path = decouple.config("DATABASE_FILE")
    try:
        with sqlite3.connect(dbase_path) as connection:
            connection.execute(
                "INSERT OR IGNORE INTO coords_t(phot_name, latitude, longitude, height) VALUES(?,?,?,?)",
                data,
            )
    except Exception as e:
        log.error(e)
    connection.close()
    log.info(
        "[%s] Added coordinates entry: Lat = %s, Long = %s, Height = %s",
        args.name,
        args.latitude,
        args.longitude,
        args.height,
    )


def cli_coords_update(args: Namespace) -> None:
    dbase_path = decouple.config("DATABASE_FILE")
    setters = list()
    data = {"name": args.name}
    if args.latitude is not None:
        setters.append("latitude = :latitude")
        data["latitude"] = args.latitude
    if args.longitude is not None:
        setters.append("longitude = :longitude")
        data["longitude"] = args.longitude
    if args.height is not None:
        setters.append("height = :height")
        data["height"] = args.height
    if not setters:
        log.error("at least one argument must be given")
        return
    sql = "UPDATE coords_t SET " + ", ".join(setters) + " WHERE phot_name = :name"
    try:
        with sqlite3.connect(dbase_path) as connection:
            connection.execute(sql, data)
    except Exception as e:
        log.error(e)
    connection.close()
    log.info("[%s] Modified coordinates entry", args.name)
    log.warning(
        "[%s] Sun/Moon data no longer valid. Delete your %s ECSV files and re-run the pipeline",
        args.name,
        args.name,
    )


def cli_coords_delete(args: Namespace) -> None:
    dbase_path = decouple.config("DATABASE_FILE")
    data = (args.name,)
    try:
        with sqlite3.connect(dbase_path) as connection:
            connection.execute("DELETE FROM coords_t WHERE phot_name = ?", data)
    except Exception as e:
        log.error(e)
    connection.close()
    log.info("[%s] Deleted coordinates entry", args.name)


def cli_coords_list(args: Namespace) -> None:
    dbase_path = decouple.config("DATABASE_FILE")
    data = (args.name,)
    try:
        with sqlite3.connect(dbase_path) as connection:
            cursor = connection.cursor()
            if args.name:
                sql = "SELECT phot_name, latitude, longitude, height FROM coords_t WHERE phot_name = ?"
                cursor.execute(sql, data)
            else:
                sql = "SELECT phot_name, latitude, longitude, height FROM coords_t ORDER BY phot_name"
                cursor.execute(sql)
            print("\n")
            paging(cursor, ("NAME", "LATITUDE", "LONGITUDE",  "HEIGHT"))
    except Exception as e:
        log.error(e)
    connection.close()


def add_args(parser: ArgumentParser) -> ArgumentParser:
    # Now parse the application specific parts
    subparser = parser.add_subparsers(dest="command")
    parser_schema = subparser.add_parser(
        "schema", help="Administrative database schema"
    )
    parser_coords = subparser.add_parser("coords", help="Coordinates Table management")

    subparser = parser_schema.add_subparsers(dest="subcommand")
    sch_cre = subparser.add_parser(  # noqa: F841
        "create", help="Create AstroPy tables and serialize to ECSV files"
    )

    subparser = parser_coords.add_subparsers(dest="subcommand")
    loc_add = subparser.add_parser("add", help="Add coordinates to a photometer")
    loc_add.add_argument(
        "-n", "--name", type=str, required=True, help="Photometer name"
    )
    loc_add.add_argument(
        "-la", "--latitude", type=float, required=True, help="Latitude [degrees]"
    )
    loc_add.add_argument(
        "-lo", "--longitude", type=float, required=True, help="Longitude [degrees]"
    )
    loc_add.add_argument(
        "-he",
        "--height",
        type=float,
        default=0.0,
        help="Height above sea level [m] (default: %(default)s)",
    )
    loc_del = subparser.add_parser("delete", help="Remove coordinates from photometer")
    loc_del.add_argument(
        "-n", "--name", type=str, required=True, help="Photometer name"
    )
    loc_del = subparser.add_parser("list", help="List photometer coordinates")
    loc_del.add_argument(
        "-n", "--name", type=str, default=None, help="Optional photometer name"
    )
    loc_upd = subparser.add_parser("update", help="Update coordinates to photometer")
    loc_upd.add_argument(
        "-n", "--name", type=str, required=True, help="Photometer name"
    )
    loc_upd.add_argument(
        "-la", "--latitude", type=float, default=None, help="Latitude [degrees]"
    )
    loc_upd.add_argument(
        "-lo", "--longitude", type=float, default=None, help="Longitude [degrees]"
    )
    loc_upd.add_argument(
        "-he",
        "--height",
        type=float,
        default=None,
        help="Height above sea level [m] (default: %(default)s)",
    )
    return parser


CMD_TABLE = {
    "schema_create": cli_schema_create,
    "coords_add": cli_coords_add,
    "coords_delete": cli_coords_delete,
    "coords_update": cli_coords_update,
    "coords_list": cli_coords_list,
}


def cli_schema(args: Namespace) -> None:
    """Create an empty able serializable to ECSV file"""
    try:
        args.subcommand
    except AttributeError:
        log.error("Missing subcommand fro command (%s)", args.command)
    else:
        func = CMD_TABLE[f"{args.command}_{args.subcommand}"]
        func(args)


def main() -> None:
    execute(
        main_func=cli_schema,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

from .api import (
    aux_table_hashes_lookup,
    aux_table_hashes_insert,
    aux_table_hashes_update,
    aux_dbase_load,
    aux_dbase_save,
    aux_table_coords_lookup,
)


__all__ = [
    "aux_dbase_load",
    "aux_dbase_save",
    "aux_table_hashes_lookup",
    "aux_table_hashes_insert",
    "aux_table_hashes_update",
    "aux_table_coords_lookup",
]



#!/bin/env python3
# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# -----------------------
# Standard Python imports
# ----------------------

import os
import sys
import logging
import logging.handlers
from datetime import datetime
import argparse
from argparse import Namespace
from typing import Union, Optional

# -------------------
# Third party imports
# -------------------

import decouple
import requests
from dateutil.relativedelta import relativedelta

# ----------------
# Module constants
# ----------------

DESCRIPTION = "Get TESS-W IDA monthly files from NextCloud server"
__version__ = "1.0.1"

OptStr = Union[str, None]

# -----------------------
# Module global variables
# -----------------------

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


def vmonth(datestr: str) -> datetime:
    return datetime.strptime(datestr, "%Y-%m")


def vyear(datestr: str) -> datetime:
    return datetime.strptime(datestr, "%Y")


def cur_month() -> datetime:
    return datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def prev_month() -> datetime:
    month = datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    return month - relativedelta(months=1)


def to_phot_dir(base_dir: Optional[str], name: str) -> str:
    cwd = os.getcwd()
    base_dir = cwd if base_dir is None else base_dir
    base_dir = os.path.join(cwd, base_dir) if not os.path.isabs(base_dir) else base_dir
    full_dir_path = os.path.join(base_dir, name)
    return full_dir_path


def makedirs(base_dir: OptStr, name: str) -> str:
    full_dir_path = to_phot_dir(base_dir, name)
    if not os.path.isdir(full_dir_path):
        os.makedirs(full_dir_path)
    return full_dir_path


def name_month(ida_file_path: str) -> tuple:
    name = os.path.basename(os.path.dirname(ida_file_path))
    month = os.path.splitext(os.path.basename(ida_file_path))[0].split("_")[1]
    return name, month


# --------------
# Work functions
# --------------


def do_ida_single_month(
    base_url: str,
    ida_base_dir: str,
    name: str,
    month: OptStr,
    exact: OptStr,
    timeout: int,
) -> None:
    target_file = name + "_" + month + ".dat" if not exact else exact
    url = os.path.join(base_url, name, target_file)
    _, month1 = name_month(target_file)
    params = None
    resp = requests.get(url, params=params, timeout=timeout)
    if resp.status_code == 404:
        log.warning("[%s] [%s] No monthly file exits: %s", name, month1, target_file)
        return
    resp.raise_for_status()  # catch other unexpected return code
    log.info("[%s] [%s] GET %s [%d OK]", name, month1, resp.url, resp.status_code)
    contents = resp.text
    if len(contents) == 0:
        log.warning("File size is 0. URL scheme may have changed ...")
    full_dir_path = makedirs(ida_base_dir, name)
    file_path = os.path.join(full_dir_path, target_file)
    with open(file_path, mode="w") as f:
        log.info("[%s] [%s] Writing %s", name, month1, file_path)
        f.write(contents)


# ================================
# COMMAND LINE INTERFACE FUNCTIONS
# ================================


def cli_ida_single(base_url: str, args: Namespace) -> None:
    month = args.month.strftime("%Y-%m") if not args.exact else None
    do_ida_single_month(
        base_url=base_url,
        ida_base_dir=args.out_dir,
        name=args.name,
        month=month,
        exact=args.exact,
        timeout=4,
    )


def cli_ida_range(base_url: str, args: Namespace) -> None:
    month = args.since
    while month <= args.until:
        do_ida_single_month(
            base_url=base_url,
            ida_base_dir=args.out_dir,
            name=args.name,
            month=month.strftime("%Y-%m"),
            exact=None,
            timeout=4,
        )
        month += relativedelta(months=1)


def cli_ida_photometers(base_url: str, args: Namespace) -> None:
    rang = sorted(args.range) if args.range is not None else None
    seq = sorted(args.list) if args.list is not None else None
    if rang:
        for i in range(rang[0], rang[1] + 1):
            args.name = "stars" + str(i)
            cli_ida_range(base_url, args)
    else:
        for i in seq:
            args.name = "stars" + str(i)
            cli_ida_range(base_url, args)


# =================
# LOGGER AND PARSER
# =================


def configure_log(args: Namespace):
    """Configure the root logger"""
    if args.verbose:
        level = logging.DEBUG
    elif args.quiet:
        level = logging.WARNING
    else:
        level = logging.INFO
    # set the root logger level
    log = logging.getLogger()
    log.setLevel(level)
    # Log formatter
    fmt = logging.Formatter("%(asctime)s [%(levelname)s] [%(name)-4s] %(message)s")
    # create console handler and set level to debug
    if args.console:
        ch = logging.StreamHandler()
        ch.setFormatter(fmt)
        ch.setLevel(logging.DEBUG)
        log.addHandler(ch)
    # Create a file handler suitable for logrotate usage
    if args.log_file:
        fh = logging.handlers.WatchedFileHandler(args.log_file)
        # fh = logging.handlers.TimedRotatingFileHandler(args.log_file, when='midnight', interval=1, backupCount=365)
        fh.setFormatter(fmt)
        fh.setLevel(logging.DEBUG)
        log.addHandler(fh)


def args_parser(name: str, version: str, description: str) -> None:
    # create the top-level parser with generic options
    parser = argparse.ArgumentParser(prog=name, description=description)
    parser.add_argument(
        "--version", action="version", version="{0} {1}".format(name, version)
    )
    parser.add_argument(
        "--console", action="store_true", help="Log to vanilla console."
    )
    parser.add_argument(
        "--log-file", type=str, metavar="<FILE>", default=None, help="Log to file."
    )
    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument("--verbose", action="store_true", help="Verbose output.")
    group1.add_argument("--quiet", action="store_true", help="Quiet output.")
    # Now parse the application specific parts
    subparser = parser.add_subparsers(dest="command")
    parser_single = subparser.add_parser(
        "single", help="Download single monthly file from a photometer"
    )
    parser_single.add_argument(
        "-n", "--name", type=str, required=True, help="Photometer name"
    )
    parser_single.add_argument(
        "-o", "--out-dir", type=str, default=None, help="Output base directory"
    )
    group1 = parser_single.add_mutually_exclusive_group(required=True)
    group1.add_argument(
        "-e", "--exact", type=str, default=None, help="Specific monthly file name"
    )
    group1.add_argument(
        "-m",
        "--month",
        type=vmonth,
        default=None,
        metavar="<YYYY-MM>",
        help="Year and Month",
    )
    parser_range = subparser.add_parser(
        "range", help="Download a month range from a photometer"
    )
    parser_range.add_argument(
        "-n", "--name", type=str, required=True, help="Photometer name"
    )
    parser_range.add_argument(
        "-s",
        "--since",
        type=vmonth,
        default=prev_month(),
        metavar="<YYYY-MM>",
        help="Year and Month (defaults to %(default)s",
    )
    parser_range.add_argument(
        "-u",
        "--until",
        type=vmonth,
        default=cur_month(),
        metavar="<YYYY-MM>",
        help="Year and Month (defaults to %(default)s",
    )
    parser_range.add_argument(
        "-o", "--out-dir", type=str, default=None, help="Output IDA base directory"
    )
    parser_phot = subparser.add_parser(
        "photometers", help="Download a month range from selected photometers"
    )
    group2 = parser_phot.add_mutually_exclusive_group(required=True)
    group2.add_argument(
        "-l",
        "--list",
        type=int,
        default=None,
        nargs="+",
        metavar="<N>",
        help="Photometer number list",
    )
    group2.add_argument(
        "-r",
        "--range",
        type=int,
        default=None,
        metavar="<N>",
        nargs=2,
        help="Photometer number range",
    )
    parser_phot.add_argument(
        "-s",
        "--since",
        type=vmonth,
        default=prev_month(),
        metavar="<YYYY-MM>",
        help="Year and Month (defaults to %(default)s",
    )
    parser_phot.add_argument(
        "-u",
        "--until",
        type=vmonth,
        default=cur_month(),
        metavar="<YYYY-MM>",
        help="Year and Month (defaults to %(default)s",
    )
    parser_phot.add_argument(
        "-o", "--out-dir", type=str, default=None, help="Output IDA base directory"
    )
    return parser


# =============
# MAIN FUNCTION
# =============


CMD_TABLE = {
    "single": cli_ida_single,
    "range": cli_ida_range,
    "photometers": cli_ida_photometers,
}


def main() -> None:
    """The main entry point specified by pyproject.toml"""
    parser = args_parser(name=__name__, version=__version__, description=DESCRIPTION)
    args = parser.parse_args(sys.argv[1:])
    configure_log(args)
    try:
        base_url = decouple.config("IDA_URL")
        func = CMD_TABLE[args.command]
        func(base_url, args)
        log.info("done!")
    except KeyboardInterrupt:
        log.warn("Application quits by user request")


if __name__ == "__main__":
    main()
